#include "gbaspc.h"
#include "armOps.h"

////////////////////////////////////////////////////////////////////////////
// Hacks
////////////////////////////////////////////////////////////////////////////

void Print(char *s) {
	__asm volatile("mov r0, %0;"
		"swi 0xff;"
		: // no ouput
		: "r" (s)
		: "r0");
}

#include "sprintf.h"
u16 lastPC;
u8 lastA;
void unimpl(u8 opcode, u16 startPC) {
//	return;
	CpuFastClear(0, VRAM, 240*160/4);
	char c[500];
	sprintf(c,"Opcode 0x%.2x broken at %.4x\nL:%.4x A:%.2x v1:%.2x v2:%.2x\n", opcode, startPC, lastPC, lastA, APU_MEM[0x30 + 0], APU_MEM[0x30 + 1]);
	DrawString(0,0,c);
	DebugGetSPCState(c);
	DrawString(0, 16, c);
//	startPC = lastPC;
	for (;;) {
		u16 pc = startPC;
		int sy = 32;
		int num = 16;
		while (num--) {
			pc += DebugGetOpcodeDesc(c, &(APU_MEM[pc]), true);
			int x = DrawString(0, sy, c);
			while (x < 240) {
				DrawChar(x, sy, ' ');
				x += 8;
			}
			sy += 8;
		}
		if (!(JOY & JOY_DOWN)) {
			startPC += DebugGetOpcodeDesc(c, &(APU_MEM[startPC]), true);
			for (int i=0; i<10; i++) WaitForBlank();
		}
		if (!(JOY & JOY_UP)) {
			startPC--;
			for (int i=0; i<10; i++) WaitForBlank();
		}
	}
	for(;;);
}

////////////////////////////////////////////////////////////////////////////
// Definitions
////////////////////////////////////////////////////////////////////////////

const u8 iplRom[64] = 
{
	0xCD,0xEF,0xBD,0xE8,0x00,0xC6,0x1D,0xD0,0xFC,0x8F,0xAA,0xF4,0x8F,0xBB,0xF5,0x78,
	0xCC,0xF4,0xD0,0xFB,0x2F,0x19,0xEB,0xF4,0xD0,0xFC,0x7E,0xF4,0xD0,0x0B,0xE4,0xF5,
	0xCB,0xF4,0xD7,0x00,0xFC,0xD0,0xF3,0xAB,0x01,0x10,0xEF,0x7E,0xF4,0x10,0xEB,0xBA,
	0xF6,0xDA,0x00,0xBA,0xF4,0xC4,0xF4,0xDD,0x5D,0xD0,0xDB,0x1F,0x00,0x00,0xC0,0xFF
};

// 8 extra bytes at the beginning of APU_MEM
// First 4 bytes stores the address of the memory write jump table
// Second 4 bytes stores the address of the memory read jump table
u8 RAW_APU_MEM[0x10000 + 8] DATA_ALIGNED_EWRAM;
u8 *APU_MEM DATA_ALIGNED_EWRAM;
u8 APU_EXTRA_MEM[0x40] DATA_ALIGNED_EWRAM;
u16 APU_PC;
u32 APU_STATE[16] DATA_ALIGNED_EWRAM;

void ApuResetTimers();
extern const u32 memWriteLUT[0x10000];
extern const u32 memReadLUT[0x10000];

// Currently 24k of IWRAM for code buffer
#define codeBufferSize 24 * 1024 / 4
// Pad by 100 in case we have overflow problems :P
u32 codeBuffer[codeBufferSize + 200];
#define ewramCodeBufferSize 32 * 1024 / 4
u32 ewramCodeBuffer[ewramCodeBufferSize + 200] DATA_ALIGNED_EWRAM;
u32 codeBufferPos;

// Page lookup structure
// NOTE: If this changes, modify apuasm.s as well
#define pageTableSize 2048
typedef struct _PageLookup {
	u32 *address;
	u16 pc;
	u16 length;
} PageLookup;
PageLookup pageTable[pageTableSize] DATA_ALIGNED_EWRAM;

u8 MakeRawPSWFromState(u32 state[16]) {
	u8 psw = 0;

	// todo: once again h bit is ignored
	psw |= ((APU_STATE[0] >> ARM_FLAG_N) & 1) << 7;
	psw |= ((APU_STATE[0] >> ARM_FLAG_V) & 1) << 6;
	psw |= ((APU_STATE[0] >> ARM_FLAG_Z) & 1) << 1;
	psw |= ((APU_STATE[0] >> ARM_FLAG_C) & 1);

	psw |= ((APU_STATE[6] >> 8) & 1) << 5;

	return psw;
}

void SetStateFromRawPSW(u32 state[16], u8 psw) {
	APU_STATE[0] = 0;

	if ((psw >> 7) & 1) APU_STATE[0] |= 1 << ARM_FLAG_N;
	if ((psw >> 6) & 1) APU_STATE[0] |= 1 << ARM_FLAG_V;
	if ((psw >> 1) & 1) APU_STATE[0] |= 1 << ARM_FLAG_Z;
	if (psw & 1) APU_STATE[0] |= 1 << ARM_FLAG_C;

	APU_STATE[6] = ((psw >> 5) & 1) << 8;
}

void ApuReset() {
	APU_MEM = RAW_APU_MEM + 8;

	// Init the ROM
	int i;
	for (i=0; i<=0x3F; i++) {
		APU_MEM[0xFFC0 + i] = iplRom[i];
		APU_EXTRA_MEM[i] = iplRom[i];
	}

	// Set up pointer to memory write and read LUT
	*((u32*)&(RAW_APU_MEM[0])) = (u32)memWriteLUT;
	*((u32*)&(RAW_APU_MEM[4])) = (u32)memReadLUT;

	// Set up the initial APU state
	APU_STATE[0] = APU_STATE[1] = APU_STATE[2] = APU_STATE[3] = 0;
	APU_STATE[4] = 0x1FF;
	APU_STATE[5] = (u32)APU_MEM;
	APU_STATE[6] = APU_STATE[8] = 0;
	APU_STATE[7] = 0xFFC0;

	CpuFastClear(0, pageTable, pageTableSize * sizeof(PageLookup) / 4);

	ApuPrepareStateAfterReload();

	codeBufferPos = 0;
}

void SetupAndEnterPageLookupAndBranch(u32 state[16]) CODE_IN_IWRAM;
u32 CompileBlock(u32 *target, u16 APU_PC);

u32 dbgNumCollisions = 0;
u32 dbgCollisionOver = 0;

void ApuDebug() {
	u32 lookup = ((1664523 * APU_STATE[7]) >> 2) & ((pageTableSize - 1) << 3);
	PageLookup *pageLookup = (PageLookup*)(((u32)pageTable) + lookup);

	pageLookup->pc = APU_STATE[7];
	pageLookup->address = &(codeBuffer[0]);
	pageLookup->length = CompileBlock(pageLookup->address, APU_STATE[7]);
	
	APU_STATE[8] = 0;
	SetupAndEnterPageLookupAndBranch(APU_STATE);
}

// Execute # cycles (note, may overflow cycles wanted to be executed).
void ApuExecute(int cycles) {
	APU_STATE[8] += cycles;

	for (;;) {
		SetupAndEnterPageLookupAndBranch(APU_STATE);

		if (((s32)APU_STATE[8]) < 0) break;

/*		PrintDisasm(0,0,19,APU_STATE[7]);
		for (;;) {
			if (!(JOY & JOY_DOWN)) {
				for (int i =0; i <10; i++) WaitForBlank();
				break;
			}
		}*/

		u32 lookup = ((1664523 * APU_STATE[7]) >> 2) & ((pageTableSize - 1) << 3);

		PageLookup *pageLookup = (PageLookup*)(((u32)pageTable) + lookup);

		/*
		// Handle collisions later
		if (pageLookup->pc != 0) {
		}
		*/

		pageLookup->pc = APU_STATE[7];
		if (codeBufferPos < codeBufferSize) {
			pageLookup->address = &(codeBuffer[codeBufferPos]);
		} else {
			pageLookup->address = &(ewramCodeBuffer[codeBufferPos - codeBufferSize]);
		}
		pageLookup->length = CompileBlock(pageLookup->address, APU_STATE[7]);

		codeBufferPos += pageLookup->length;
		if (codeBufferPos > ewramCodeBufferSize + codeBufferSize) {
			CpuFastClear(0, pageTable, pageTableSize * sizeof(PageLookup) / 4);
			codeBufferPos = 0;
/*			CpuFastClear(0, VRAM, 240*160/4);
			char c[50];
			DrawString(0, 0, "Too many opcodes!");
			sprintf(c, "%d collisions, %d", dbgNumCollisions, dbgCollisionOver);
			DrawString(0, 8, c);
			for(;;);*/
		}
	}
}

////////////////////////////////////////////////////////////////////////////
// Memory handlers
////////////////////////////////////////////////////////////////////////////

// Memory handlers use R3 and R4 as temporaries.
// R1 is the address to check
#define EmitMemReadCheckR1(reg) \
	*target++ = LDRN_I(APU_RAMBASE, ARM_R3, 4); \
	*target++ = LDR_RS(ARM_R3, ARM_R3, 2, SHIFT_LSL, ARM_R1); \
	*target++ = MOV_RI(ARM_R4, 0, ARM_PC, 0, 0); \
	*target++ = BX(ARM_R3)

// R0 is the value that was written to memory - can be killed
// R1 is the address to check (MUST NOT BE MODIFIED)
// R2 MUST NOT BE MODIFIED
// R3 is temporary
// R4 is the jumpback address
#define EmitMemWriteCheckR1() \
	*target++ = LDRN_I(APU_RAMBASE, ARM_R3, 8); \
	*target++ = LDR_RS(ARM_R3, ARM_R3, 2, SHIFT_LSL, ARM_R1); \
	*target++ = MOV_RI(ARM_R4, 0, ARM_PC, 0, 0); \
	*target++ = BX(ARM_R3)

#define EmitMemWriteCheckR1Reg(reg) \
	*target++ = AND_I(ARM_R0, (reg), 0, 0xff); \
	EmitMemWriteCheckR1()

////////////////////////////////////////////////////////////////////////////
// Program status word macros
////////////////////////////////////////////////////////////////////////////

// Pack returns result in r0, unpack expects arg in r0

#define ARM_FLAG_N 31
#define ARM_FLAG_Z 30
#define ARM_FLAG_C 29
#define ARM_FLAG_V 28

// Builds flag byte from ARM flags - DP appears to be ignored by Snes9x sound code.
// This scares me.
#define PACK_PSW() \
	*target++ = MRS(ARM_R1); \
	*target++ = MOV_I(ARM_R3, 0, 0, 1); \
	*target++ = AND_RI(ARM_R2, ARM_R3, ARM_R1, SHIFT_LSR, ARM_FLAG_N); \
	*target++ = ORR_RI(ARM_R0, ARM_R0, ARM_R2, SHIFT_LSL, 7); \
	*target++ = AND_RI(ARM_R2, ARM_R3, ARM_R1, SHIFT_LSR, ARM_FLAG_V); \
	*target++ = ORR_RI(ARM_R0, ARM_R0, ARM_R2, SHIFT_LSL, 6); \
	*target++ = AND_RI(ARM_R2, ARM_R3, ARM_R1, SHIFT_LSR, ARM_FLAG_Z); \
	*target++ = ORR_RI(ARM_R0, ARM_R0, ARM_R2, SHIFT_LSL, 1); \
	*target++ = AND_RI(ARM_R2, ARM_R3, ARM_R1, SHIFT_LSR, ARM_FLAG_C); \
	*target++ = ORR_RI(ARM_R0, ARM_R0, ARM_R2, 0, 0)

#define UNPACK_PSW() \
	*target++ = MOV_I(ARM_R3, 0, 0, 1); \
	*target++ = MOV_I(ARM_R1, 0, 0, 0); \
	*target++ = AND_RI(ARM_R2, ARM_R3, ARM_R0, SHIFT_LSR, 7); \
	*target++ = ORR_RI(ARM_R1, ARM_R1, ARM_R2, SHIFT_LSL, ARM_FLAG_N); \
	*target++ = AND_RI(ARM_R2, ARM_R3, ARM_R0, SHIFT_LSR, 6); \
	*target++ = ORR_RI(ARM_R1, ARM_R1, ARM_R2, SHIFT_LSL, ARM_FLAG_V); \
	*target++ = AND_RI(ARM_R2, ARM_R3, ARM_R0, SHIFT_LSR, 1); \
	*target++ = ORR_RI(ARM_R1, ARM_R1, ARM_R2, SHIFT_LSL, ARM_FLAG_Z); \
	*target++ = AND_RI(ARM_R2, ARM_R3, ARM_R0, 0, 0); \
	*target++ = ORR_RI(ARM_R1, ARM_R1, ARM_R2, SHIFT_LSL, ARM_FLAG_C); \
	*target++ = MSR_R(ARM_R1)

////////////////////////////////////////////////////////////////////////////
// Macros to loads addresses into R1
////////////////////////////////////////////////////////////////////////////
#define LOADDP() \
	*target++ = ADD_I(ARM_R1, APU_DP, 0, APU_MEM[APU_PC++])

#define LOADDP_OFS(reg) \
	*target++ = ADD_I(ARM_R1, APU_DP, 0, APU_MEM[APU_PC++]); \
	*target++ = ADD_RI(ARM_R1, ARM_R1, (reg), SHIFT_LSR ,24)

#define LOADDP_X() \
	LOADDP_OFS(APU_REG_X); \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemReadCheckR1(ARM_R0); \
	*target++ = ADD_I(ARM_R1, ARM_R1, 0, 1); \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R2, ARM_R1); \
	EmitMemReadCheckR1(ARM_R2); \
	*target++ = ORR_RI(ARM_R1, ARM_R0, ARM_R2, SHIFT_LSL, 8)

// Clear the 0x00ff0000 bits from R1, incase we have some spillover bits
#define LOADDP_Y() \
	LOADDP(); \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemReadCheckR1(ARM_R0); \
	*target++ = ADD_I(ARM_R1, ARM_R1, 0, 1); \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R2, ARM_R1); \
	EmitMemReadCheckR1(ARM_R2); \
	*target++ = ORR_RI(ARM_R1, ARM_R0, ARM_R2, SHIFT_LSL, 8); \
	*target++ = ADD_RI(ARM_R1, ARM_R1, APU_REG_Y, SHIFT_LSR, 24); \
	*target++ = BIC_I(ARM_R1, ARM_R1, 8, 0xff)

#define LOADIMM() \
	*target++ = MOV_I(ARM_R1, 0, 0, APU_MEM[APU_PC++]); \
	*target++ = ORR_I(ARM_R1, ARM_R1, 12, APU_MEM[APU_PC++])

// Clear the 0x00ff0000 bits from R1, incase we have some spillover bits
#define LOADIMM_OFS(reg) \
	LOADIMM(); \
	*target++ = ADD_RI(ARM_R1, ARM_R1, (reg), SHIFT_LSR, 24); \
	*target++ = BIC_I(ARM_R1, ARM_R1, 8, 0xff)

// Don't need a read check on the second LRDB because it will be written to
#define LOADXY() \
	*target++ = MOV_RI(ARM_R1, 0, APU_REG_Y, SHIFT_LSR ,24); \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R2, ARM_R1); \
	EmitMemReadCheckR1(ARM_R2); \
	*target++ = MOV_RI(ARM_R2, 0, ARM_R2, SHIFT_LSL, 24); \
	*target++ = MOV_RI(ARM_R1, 0, APU_REG_X, SHIFT_LSR ,24); \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSL, 24)

// Don't need a read check on the second LRDB because it will be written to
#define LOADDPDP() \
	LOADDP(); \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R2, ARM_R1); \
	EmitMemReadCheckR1(ARM_R2); \
	*target++ = MOV_RI(ARM_R2, 0, ARM_R2, SHIFT_LSL, 24); \
	LOADDP(); \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSL, 24)

#define LOADDPIMM() \
	*target++ = MOV_I(ARM_R2, 0, 4, APU_MEM[APU_PC++]); \
	LOADDP(); \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemReadCheckR1(ARM_R0); \
	*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSL, 24)

#define PUSH_BYTE(reg) \
	*target++ = STRB_R(APU_RAMBASE, (reg), APU_REG_SP); \
	*target++ = SUB_I(APU_REG_SP, APU_REG_SP, 0, 1); \
	*target++ = BIC_I(APU_REG_SP, APU_REG_SP, 12, 0xff); \
	*target++ = ORR_I(APU_REG_SP, APU_REG_SP, 12, 1)

#define POP_BYTE(reg) \
	*target++ = ADD_I(APU_REG_SP, APU_REG_SP, 0, 1); \
	*target++ = BIC_I(APU_REG_SP, APU_REG_SP, 12, 0xff); \
	*target++ = ORR_I(APU_REG_SP, APU_REG_SP, 12, 1); \
	*target++ = LDRB_R(APU_RAMBASE, (reg), APU_REG_SP)

#define PUSH_WORD(reg) \
	*target++ = MOV_RI(ARM_R0, 0, (reg), SHIFT_LSR, 8); \
	PUSH_BYTE(ARM_R0); \
	PUSH_BYTE((reg))

#define POP_WORD(reg) \
	POP_BYTE((reg)); \
	POP_BYTE(ARM_R0); \
	*target++ = ORR_RI((reg), (reg), ARM_R0, SHIFT_LSL, 8)

////////////////////////////////////////////////////////////////////////////
// Convienience macros
////////////////////////////////////////////////////////////////////////////
#define LOADR1_MOVR(reg) \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemReadCheckR1(ARM_R0); \
	*target++ = MOV_RI((reg), 0, ARM_R0, SHIFT_LSL, 24); \
	*target++ = MOVS_RI((reg), 0, (reg), 0, 0)

#define STORER1(reg) \
	*target++ = MOV_RI(ARM_R0, 0, (reg), SHIFT_LSR, 24); \
	*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemWriteCheckR1()

#define ADC_FROMR1() \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemReadCheckR1(ARM_R0); \
	*target++ = ADCS_RI(APU_REG_A, APU_REG_A, ARM_R0, SHIFT_LSL, 24)

#define SBC_FROMR1() \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemReadCheckR1(ARM_R0); \
	*target++ = SBCS_RI(APU_REG_A, APU_REG_A, ARM_R0, SHIFT_LSL, 24)

// V flag is bit 28
#define CMP_FROMIMM(reg) \
	*target++ = MRS(ARM_R0); \
	*target++ = AND_I(ARM_R0, ARM_R0, 4, 0x10); \
	*target++ = CMPS_I(0, (reg), 4, APU_MEM[APU_PC++]); \
	*target++ = MRS(ARM_R2); \
	*target++ = BIC_I(ARM_R2, ARM_R2, 4, 0x10); \
	*target++ = ORR_RI(ARM_R2, ARM_R2, ARM_R0, 0, 0); \
	*target++ = MSR_R(ARM_R2)

#define CMP_FROMR1(reg) \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R2, ARM_R1); \
	EmitMemReadCheckR1(ARM_R2); \
	*target++ = MRS(ARM_R0); \
	*target++ = AND_I(ARM_R0, ARM_R0, 4, 0x10); \
	*target++ = CMPS_RI(0, (reg), ARM_R2, SHIFT_LSL, 24); \
	*target++ = MRS(ARM_R2); \
	*target++ = BIC_I(ARM_R2, ARM_R2, 4, 0x10); \
	*target++ = ORR_RI(ARM_R2, ARM_R2, ARM_R0, 0, 0); \
	*target++ = MSR_R(ARM_R2)

#define CMP_R0_R2(reg) \
	*target++ = MRS(ARM_R1); \
	*target++ = AND_I(ARM_R1, ARM_R1, 4, 0x10); \
	*target++ = CMPS_RI(0, ARM_R0, ARM_R2, 0, 0); \
	*target++ = MRS(ARM_R2); \
	*target++ = BIC_I(ARM_R2, ARM_R2, 4, 0x10); \
	*target++ = ORR_RI(ARM_R2, ARM_R2, ARM_R1, 0, 0); \
	*target++ = MSR_R(ARM_R2)

#define AND_FROMR1() \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemReadCheckR1(ARM_R0); \
	*target++ = AND_RI(APU_REG_A, APU_REG_A, ARM_R0, SHIFT_LSL, 24); \
	*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, 0, 0)

#define OR_FROMR1() \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemReadCheckR1(ARM_R0); \
	*target++ = ORR_RI(APU_REG_A, APU_REG_A, ARM_R0, SHIFT_LSL, 24); \
	*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, 0, 0)

#define EOR_FROMR1() \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemReadCheckR1(ARM_R0); \
	*target++ = EOR_RI(APU_REG_A, APU_REG_A, ARM_R0, SHIFT_LSL, 24); \
	*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, 0, 0)

#define INC_FROMR1() \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSL, 24); \
	*target++ = ADD_I(ARM_R0, ARM_R0, 4, 1); \
	*target++ = MOVS_RI(ARM_R0, 0, ARM_R0, 0, 0); \
	*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24); \
	*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemWriteCheckR1()	

#define DEC_FROMR1() \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSL, 24); \
	*target++ = SUB_I(ARM_R0, ARM_R0, 4, 1); \
	*target++ = MOVS_RI(ARM_R0, 0, ARM_R0, 0, 0); \
	*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24); \
	*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemWriteCheckR1()

#define SET_DP_BIT(bit) \
	LOADDP(); \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	*target++ = ORR_I(ARM_R0, ARM_R0, 0, 1 << (bit)); \
	*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemWriteCheckR1(); \
	cycles += 4

#define CLEAR_DP_BIT(bit) \
	LOADDP(); \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	*target++ = BIC_I(ARM_R0, ARM_R0, 0, 1 << (bit)); \
	*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemWriteCheckR1(); \
	cycles += 4

#define ASL_FROMR1() \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	*target++ = MOVS_RI(ARM_R0, 0, ARM_R0, SHIFT_LSL, 25); \
	*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24); \
	*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemWriteCheckR1()

#define LSR_FROMR1() \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	*target++ = MOVS_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 1); \
	*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemWriteCheckR1()

// This is somewhat confusing, so I will explain it to myself:
// ORR with 1 ROR 8 if carry flag is set
// Shift left by 25, to put high bit in carry flag
// Then set N and Z with the MOV, C flag will remain unchanged
#define ROL_FROMR1() \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	*target++ = MOVS_RI(ARM_R0, 0, ARM_R0, SHIFT_LSL, 25); \
	*target++ = ORRCS_I(ARM_R0, ARM_R0, 4, 1); \
	*target++ = MOVS_RI(ARM_R0, 0, ARM_R0, 0, 0); \
	*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24); \
	*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemWriteCheckR1()

#define ROR_FROMR1() \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	*target++ = MOVS_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 1); \
	*target++ = ORRCS_I(ARM_R0, ARM_R0, 0, 0x80); \
	*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSL, 24); \
	*target++ = MOVS_RI(ARM_R0, 0, ARM_R0, 0, 0); \
	*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24); \
	EmitMemWriteCheckR1()

#define BRANCH_REL(type) \
	val = ((s8*)APU_MEM)[APU_PC++]; \
	*target++ = ADD_I(APU_REG_PC, APU_REG_PC, 0, APU_PC - startPC); \
	if (val < 0) *target++ = SUB_I(APU_REG_PC, APU_REG_PC, 0, (u8)(-val)); \
	else *target++ = ADD_I(APU_REG_PC, APU_REG_PC, 0, val); \
	*target++ = MOV_I(ARM_R3, 0, 0, cycles); \
	*target++ = MOV##type(ARM_PC, 0, ARM_LR, 0, 0); \
	if (val < 0) *target++ = ADD_I(APU_REG_PC, APU_REG_PC, 0, (u8)(-val)); \
	else *target++ = SUB_I(APU_REG_PC, APU_REG_PC, 0, val); \
	*target++ = SUB_I(ARM_R3, ARM_R3, 0, 2); \
	*target++ = MOV_RI(ARM_PC, 0, ARM_LR, 0, 0); \
	goto doneRecomp

#define BRANCH_DPBIT_SET(bit) \
	cycles += 7; \
	LOADDP(); \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemReadCheckR1(ARM_R0); \
	*target++ = MRS(ARM_R4); \
	*target++ = TSTS_I(0, ARM_R0, 0, 1 << (bit)); \
	BRANCH_REL(NE_RI)

#define BRANCH_DPBIT_CLR(bit) \
	cycles += 7; \
	LOADDP(); \
	*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); \
	EmitMemReadCheckR1(ARM_R0); \
	*target++ = MRS(ARM_R4); \
	*target++ = TSTS_I(0, ARM_R0, 0, 1 << (bit)); \
	BRANCH_REL(EQ_RI)

#define TCALL(value) \
	cycles += 8; \
	unimpl(0x01, startPC)
/*	TODO: I think this works, but I'm not sure, and want to test if it's breaking things
	cycles += 6; \
	*target++ = MRS(ARM_R4); \
	*target++ = B(3); \
	*target++ = 0xFFC0 + ((15 - value) * 2); \
	*target++ = APU_PC; \
	*target++ = SuccessJumpAddr; \
	*target++ = cycles; \
	*target++ = LDRN_I(ARM_PC, ARM_R3, 5*4); \
	PUSH_WORD(ARM_R3); \
	*target++ = LDRN_I(ARM_PC, ARM_R3, 12*4); \
	*target++ = LDRH_R(APU_RAMBASE, ARM_R3, ARM_R3); \
	*target++ = LDRN_I(ARM_PC, ARM_R0, 12*4); \
	*target++ = LDRN_I(ARM_PC, ARM_R1, 12*4); \
	*target++ = SUB_RI(ARM_R12, ARM_R12, ARM_R1, 0, 0); \
	*target++ = BX(ARM_R0); \
	goto doneRecomp*/

////////////////////////////////////////////////////////////////////////////
// Dynamic recompilation entry function, call with a target
// memory block to recompile all instructions until a jump or sleep is hit.
//
// R0, R2 are temporaries
// R1 is the register in which all memory checks are done.
// R3, R4 is used by the memory checker.
// R3 is also used as the ARM_PC storage when exiting
// R4 is also used as the flag storage when exiting
// R5-R12 are reserved for execution (check apu.h for the uses)
// R13 is GBA sp, R14 is lr, R15 is PC
////////////////////////////////////////////////////////////////////////////
u32 CompileBlock(u32 *target, u16 APU_PC) {
	int cycles = 0;
	u16 startPC = APU_PC;
	u32 *startTarget = target;
	s8 val;

	for (;;) {
		switch (APU_MEM[APU_PC++]) {
			////////////////////////////////////////////////////////////////////////////
			// 8 bit load commands
			////////////////////////////////////////////////////////////////////////////
			
			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    A, #inm    E8    2     2     A <- inm                  N......Z
			case 0xE8:
				*target++ = MOV_I(APU_REG_A, 0, 4, APU_MEM[APU_PC++]);
				*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    A, (X)     E6    1     3     A <- (X)                  N......Z
			case 0xE6:
				*target++ = MOV_RI(ARM_R1, 0, APU_REG_X, SHIFT_LSR, 24);
				LOADR1_MOVR(APU_REG_A);
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    A, (X)+    BF    1     4     A <- (X) with auto inc    N......Z
			case 0xBF:
				*target++ = MOV_RI(ARM_R1, 0, APU_REG_X, SHIFT_LSR, 24);
				*target++ = ADD_I(APU_REG_X, APU_REG_X, 4, 1);
				LOADR1_MOVR(APU_REG_A);
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    A, dp      E4    2     3     A <- (dp)                 N......Z
			case 0xE4:
				LOADDP();
				LOADR1_MOVR(APU_REG_A);
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    A, dp+X    F4    2     4     A <- (dp+X)               N......Z
			case 0xF4:
				LOADDP_OFS(APU_REG_X);
				LOADR1_MOVR(APU_REG_A);
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    A, labs    E5    3     4     A <- (abs)                N......Z
			case 0xE5:
				LOADIMM();
				// TODO: Can optimize this read because it is a constant.
				LOADR1_MOVR(APU_REG_A);
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    A, labs+X  F5    3     5     A <- (abs+X)              N......Z
			case 0xF5:
				LOADIMM_OFS(APU_REG_X);
				LOADR1_MOVR(APU_REG_A);
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    A, labs+Y  F6    3     5     A <- (abs+Y)              N......Z
			case 0xF6:
				LOADIMM_OFS(APU_REG_Y);
				LOADR1_MOVR(APU_REG_A);
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    A, (dp+X)  E7    2     6     A <- ((dp+X+1)(dp+X))     N......Z
			case 0xE7:
				LOADDP_X();
				LOADR1_MOVR(APU_REG_A);
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    A, (dp)+Y  F7    2     6     A <- ((dp+1)(dp)+Y)       N......Z
			case 0xF7:
				LOADDP_Y();
				LOADR1_MOVR(APU_REG_A);
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    X, #inm    CD    2     2     X <- inm                  N......Z
			case 0xCD:
				*target++ = MOV_I(APU_REG_X, 0, 4, APU_MEM[APU_PC++]);
				*target++ = MOVS_RI(APU_REG_X, 0, APU_REG_X, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    X, dp      F8    2     3     X <- (dp)                 N......Z
			case 0xF8:
				LOADDP();
				LOADR1_MOVR(APU_REG_X);
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    X, dp+Y    F9    2     4     X <- (dp+Y)               N......Z
			case 0xF9:
				LOADDP_OFS(APU_REG_Y);
				LOADR1_MOVR(APU_REG_X);
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    X, labs    E9    3     4     X <- (abs)                N......Z
			case 0xE9:
				LOADIMM();
				// TODO: Can optimize this read because it is a constant
				LOADR1_MOVR(APU_REG_X);
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    Y, #inm    8D    2     2     Y <- inm                  N......Z
			case 0x8D:
				*target++ = MOV_I(APU_REG_Y, 0, 4, APU_MEM[APU_PC++]);
				*target++ = MOVS_RI(APU_REG_Y, 0, APU_REG_Y, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    Y, dp      EB    2     3     Y <- (dp)                 N......Z
			case 0xEB:
				LOADDP();
				LOADR1_MOVR(APU_REG_Y);
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    Y, dp+X    FB    2     4     Y <- (dp+X)               N......Z
			case 0xFB:
				LOADDP_OFS(APU_REG_X);
				LOADR1_MOVR(APU_REG_Y);
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    Y, labs    EC    3     4     Y <- (abs)                N......Z
			case 0xEC:
				LOADIMM();
				LOADR1_MOVR(APU_REG_Y);
				cycles += 4;
				break;

			////////////////////////////////////////////////////////////////////////////
			// 8 bit store commands
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    (X),A      C6    1     4     A -> (X)                  ........
			case 0xC6:
				*target++ = MOV_RI(ARM_R1, 0, APU_REG_X, SHIFT_LSR, 24);
				STORER1(APU_REG_A);
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    (X)+,A     AF    1     4     A -> (X) with auto inc    ........
			case 0xAF:
				*target++ = MOV_RI(ARM_R1, 0, APU_REG_X, SHIFT_LSR, 24);
				*target++ = ADD_I(APU_REG_X, APU_REG_X, 4, 1);
				STORER1(APU_REG_A);
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    dp,A       C4    2     4     A -> (dp)                 ........
			case 0xC4:
				LOADDP();
				STORER1(APU_REG_A);
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    dp+X,A     D4    2     5     A -> (dp+X)               ........
			case 0xD4:
				LOADDP_OFS(APU_REG_X);
				STORER1(APU_REG_A);
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    labs,A     C5    3     5     A -> (abs)                ........
			case 0xC5:
				LOADIMM();
				STORER1(APU_REG_A);
				cycles += 5;
				break;
				
			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    labs+X,A   D5    3     6     A -> (abs+X)              ........
			case 0xD5:
				LOADIMM_OFS(APU_REG_X);
				STORER1(APU_REG_A);
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    labs+Y,A   D6    3     6     A -> (abs+Y)              ........
			case 0xD6:
				LOADIMM_OFS(APU_REG_Y);
				STORER1(APU_REG_A);
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    (dp+X),A   C7    2     7     A -> ((dp+X+1)(dp+X))     ........
			case 0xC7:
				LOADDP_X();
				STORER1(APU_REG_A);
				cycles += 7;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    (dp)+Y,A   D7    2     7     A -> ((dp+1)(dp)+Y)       ........
			case 0xD7:
				LOADDP_Y();
				STORER1(APU_REG_A);
				cycles += 7;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    dp,X       D8    2     4     X -> (dp)                 ........
			case 0xD8:
				LOADDP();
				STORER1(APU_REG_X);
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    dp+Y,X     D9    2     5     X -> (dp+Y)               ........
			case 0xD9:
				LOADDP_OFS(APU_REG_Y);
				STORER1(APU_REG_X);
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    labs,X     C9    3     5     X -> (abs)                ........
			case 0xC9:
				LOADIMM();
				STORER1(APU_REG_X);
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    dp,Y       CB    2     4     Y -> (dp)                 ........
			case 0xCB:
				LOADDP();
				STORER1(APU_REG_Y);
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    dp+X,Y     DB    2     5     Y -> (dp+X)               ........
			case 0xDB:
				LOADDP_OFS(APU_REG_X);
				STORER1(APU_REG_Y);
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    labs,Y     0C    3     5     Y -> (abs)                ........
			case 0x0C:
				LOADIMM();
				STORER1(APU_REG_Y);
				cycles += 5;
				break;

			////////////////////////////////////////////////////////////////////////////
			// 8 bit misc. commands
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    A, X        7D    1     2     A <- X                   N......Z
			case 0x7D:
				*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_X, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    A, Y        DD    1     2     A <- Y                   N......Z
			case 0xDD:
				*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_Y, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    X, A        5D    1     2     X <- A                   N......Z
			case 0x5D:
				*target++ = MOVS_RI(APU_REG_X, 0, APU_REG_A, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    Y, A        FD    1     2     Y <- A                   N......Z
			case 0xFD:
				*target++ = MOVS_RI(APU_REG_Y, 0, APU_REG_A, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    X, SP       9D    1     2     X <- SP                  N......Z
			case 0x9D:
				*target++ = MOV_RI(APU_REG_X, 0, APU_REG_SP, SHIFT_LSL, 24);
				*target++ = MOVS_RI(APU_REG_X, 0, APU_REG_X, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    SP, X       BD    1     2    SP <- X                   ........
			case 0xBD:
				*target++ = BIC_I(APU_REG_SP, APU_REG_SP, 0, 0xFF);
				*target++ = ORR_RI(APU_REG_SP, APU_REG_SP, APU_REG_X, SHIFT_LSR, 24);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    dp(d),dp(s) FA    3     5    (dp(d)) <- (dp(s))        ........
			case 0xFA:
				LOADDP();
				*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemReadCheckR1(ARM_R0);
				LOADDP();
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   MOV    dp,#inm     8F    3     5    (dp) <- inm               ........
			case 0x8F:
				*target++ = MOV_I(ARM_R0, 0, 0, APU_MEM[APU_PC++]);
				LOADDP();
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 5;
				break;

			////////////////////////////////////////////////////////////////////////////
			// Add with carry commands
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   ADC    A,#inm      88    2     2     A <- A+inm+C            NV..H..ZC
			case 0x88:
				*target++ = ADCS_I(APU_REG_A, APU_REG_A, 4, APU_MEM[APU_PC++]);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   ADC    A,(X)       86    1     3     A <- A+(X)+C            NV..H..ZC
			case 0x86:
				*target++ = MOV_RI(ARM_R1, 0, APU_REG_X, SHIFT_LSR, 24);
				ADC_FROMR1();
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   ADC    A,dp        84    2     3     A <- A+(dp)+C           NV..H..ZC
			case 0x84:
				LOADDP();
				ADC_FROMR1();
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   ADC    A,dp+X      94    2     4     A <- A+(dp+X)+C         NV..H..ZC
			case 0x94:
				LOADDP_OFS(APU_REG_X);
				ADC_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   ADC    A,labs      85    3     4     A <- A+(abs)+C          NV..H..ZC
			case 0x85:
				LOADIMM();
				ADC_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   ADC    A,labs+X    95    3     5     A <- A+(abs+X)+C        NV..H..ZC
			case 0x95:
				LOADIMM_OFS(APU_REG_X);
				ADC_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   ADC    A,labs+Y    96    3     5     A <- A+(abs+Y)+C        NV..H..ZC
			case 0x96:
				LOADIMM_OFS(APU_REG_Y);
				ADC_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   ADC    A,(dp+X)    87    2     6     A <- A+((dp+X+1)(dp+X)) NV..H..ZC
			case 0x87:
				LOADDP_X();
				ADC_FROMR1();
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   ADC    A,(dp)+Y    97    2     6     A <- A+((dp+1)(dp)+Y)   NV..H..ZC
			case 0x97:
				LOADDP_Y();
				ADC_FROMR1();
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   ADC    (X),(Y)     99    1     5   (X) <- (X)+(Y)+C          NV..H..ZC
			case 0x99:
				LOADXY();
				*target++ = ADCS_RI(ARM_R0, ARM_R0, ARM_R2, 0, 0);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   ADC    dp(d),dp(s) 89    3     6 (dp(d))<-(dp(d))+(dp(s))+C  NV..H..ZC
			case 0x89:
				LOADDPDP();
				*target++ = ADCS_RI(ARM_R0, ARM_R0, ARM_R2, 0, 0);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   ADC    dp,#inm     98    3     5  (dp) <- (dp)+inm+C         NV..H..ZC
			case 0x98:
				LOADDPIMM();
				*target++ = ADCS_RI(ARM_R0, ARM_R0, ARM_R2, 0, 0);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 5;
				break;

			////////////////////////////////////////////////////////////////////////////
			// Subtract with carry commands
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   SBC    A,#inm      A8    2     2     A <- A-inm-!C           NV..H..ZC
			case 0xA8:
				*target++ = SBCS_I(APU_REG_A, APU_REG_A, 4, APU_MEM[APU_PC++]);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   SBC    A,(X)       A6    1     3     A <- A-(X)-!C           NV..H..ZC
			case 0xA6:
				*target++ = MOV_RI(ARM_R1, 0, APU_REG_X, SHIFT_LSR, 24);
				SBC_FROMR1();
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   SBC    A,dp        A4    2     3     A <- A-(dp)-!C          NV..H..ZC
			case 0xA4:
				LOADDP();
				SBC_FROMR1();
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   SBC    A,dp+X      B4    2     4     A <- A-(dp+X)-!C        NV..H..ZC
			case 0xB4:
				LOADDP_OFS(APU_REG_X);
				SBC_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   SBC    A,labs      A5    3     4     A <- A-(abs)-!C         NV..H..ZC
			case 0xA5:
				LOADIMM();
				SBC_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   SBC    A,labs+X    B5    3     5     A <- A-(abs+X)-!C       NV..H..ZC
			case 0xB5:
				LOADIMM_OFS(APU_REG_X);
				SBC_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   SBC    A,labs+Y    B6    3     5     A <- A-(abs+Y)-!C       NV..H..ZC
			case 0xB6:
				LOADIMM_OFS(APU_REG_Y);
				SBC_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   SBC    A,(dp+X)    A7    2     6  A <- A-((dp+X+1)(dp+X))-!C NV..H..ZC
			case 0xA7:
				LOADDP_X();
				SBC_FROMR1();
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   SBC    A,(dp)+Y    B7    2     6  A <- A-((dp+1)(dp)+Y)-!C   NV..H..ZC
			case 0xB7:
				LOADDP_Y();
				SBC_FROMR1();
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   SBC    (X),(Y)     B9    1     5   (X) <- (X)-(Y)-!C         NV..H..ZC
			case 0xB9:
				LOADXY();
				*target++ = SBCS_RI(ARM_R0, ARM_R0, ARM_R2, 0, 0);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   SBC    dp(d),dp(s) A9    3     6 (dp(d))<-(dp(d))-(dp(s))-!C NV..H..ZC
			case 0xA9:
				LOADDPDP();
				*target++ = SBCS_RI(ARM_R0, ARM_R0, ARM_R2, 0, 0);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   SBC    dp,#inm     B8    3     5  (dp) <- (dp)-inm-!C        NV..H..ZC
			case 0xB8:
				LOADDPIMM();
				*target++ = SBCS_RI(ARM_R0, ARM_R0, ARM_R2, 0, 0);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 5;
				break;

			////////////////////////////////////////////////////////////////////////////
			// Comparison commands
			////////////////////////////////////////////////////////////////////////////


			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    A,#inm      68    2     2     A-inm                   N......ZC
			case 0x68:
				CMP_FROMIMM(APU_REG_A);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    A,(X)       66    1     3     A-(X)                   N......ZC
			case 0x66:
				*target++ = MOV_RI(ARM_R1, 0, APU_REG_X, SHIFT_LSR, 24);
				CMP_FROMR1(APU_REG_A);
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    A,dp        64    2     3     A-(dp)                  N......ZC
			case 0x64:
				LOADDP();
				CMP_FROMR1(APU_REG_A);
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    A,dp+X      74    2     4     A-(dp+X)                N......ZC
			case 0x74:
				LOADDP_OFS(APU_REG_X);
				CMP_FROMR1(APU_REG_A);
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    A,labs      65    3     4     A-(abs)                 N......ZC
			case 0x65:
				LOADIMM();
				CMP_FROMR1(APU_REG_A);
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    A,labs+X    75    3     5     A-(abs+X)               N......ZC
			case 0x75:
				LOADIMM_OFS(APU_REG_X);
				CMP_FROMR1(APU_REG_A);
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    A,labs+Y    76    3     5     A-(abs+Y)               N......ZC
			case 0x76:
				LOADIMM_OFS(APU_REG_Y);
				CMP_FROMR1(APU_REG_A);
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    A,(dp+X)    67    2     6     A-((dp+X+1)(dp+X))      N......ZC
			case 0x67:
				LOADDP_X();
				CMP_FROMR1(APU_REG_A);
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    A,(dp)+Y    77    2     6     A-((dp+1)(dp)+Y)        N......ZC
			case 0x77:
				LOADDP_Y();
				CMP_FROMR1(APU_REG_A);
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    (X),(Y)     79    1     5     (X)-(Y)                 N......ZC
			case 0x79:
				LOADXY();
				CMP_R0_R2();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    dp(d),dp(s) 69    3     6     (dp(d))-(dp(s))         N......ZC
			case 0x69:
				LOADDPDP();
				CMP_R0_R2();
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    dp,#inm     78    3     5     (dp)-inm                N......ZC
			case 0x78:
				LOADDPIMM();
				CMP_R0_R2();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    X,#inm      C8    2     2     X-inm                   N......ZC
			case 0xC8:
				CMP_FROMIMM(APU_REG_X);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    X,dp        3E    2     3     X-(dp)                  N......ZC
			case 0x3E:
				LOADDP();
				CMP_FROMR1(APU_REG_X);
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    X,labs      1E    3     4     X-(abs)                 N......ZC
			case 0x1E:
				LOADIMM();
				CMP_FROMR1(APU_REG_X);
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    Y,#inm      AD    2     2     Y-inm                   N......ZC
			case 0xAD:
				CMP_FROMIMM(APU_REG_Y);
				cycles += 2;				
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    Y,dp        7E    2     3     Y-(dp)                  N......ZC
			case 0x7E:
				LOADDP();
				CMP_FROMR1(APU_REG_Y);
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CMP    Y,labs      5E    3     4     Y-(abs)                 N......ZC
			case 0x5E:
				LOADIMM();
				CMP_FROMR1(APU_REG_Y);
				cycles += 4;
				break;

			////////////////////////////////////////////////////////////////////////////
			// Logical AND commands
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   AND    A,#inm      28    2     2     A <- A AND inm           N......Z.
			case 0x28:
				*target++ = AND_I(APU_REG_A, APU_REG_A, 4, APU_MEM[APU_PC++]);
				*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   AND    A,(X)       26    1     3     A <- A AND (X)           N......Z.
			case 0x26:
				*target++ = MOV_RI(ARM_R1, 0, APU_REG_X, SHIFT_LSR, 24);
				AND_FROMR1();
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   AND    A,dp        24    2     3     A <- A AND (dp)          N......Z.
			case 0x24:
				LOADDP();
				AND_FROMR1();
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   AND    A,dp+X      34    2     4     A <- A AND (dp+X)        N......Z.
			case 0x34:
				LOADDP_OFS(APU_REG_X);
				AND_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   AND    A,labs      25    3     4     A <- A AND (abs)         N......Z.
			case 0x25:
				LOADIMM();
				AND_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   AND    A,labs+X    35    3     5     A <- A AND (abs+X)       N......Z.
			case 0x35:
				LOADIMM_OFS(APU_REG_X);
				AND_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   AND    A,labs+Y    36    3     5     A <- A AND (abs+Y)       N......Z.
			case 0x36:
				LOADIMM_OFS(APU_REG_Y);
				AND_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   AND    A,(dp+X)    27    2     6  A <- A AND ((dp+X+1)(dp+X)) N......Z.
			case 0x27:
				LOADDP_X();
				AND_FROMR1();
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   AND    A,(dp)+Y    37    2     6   A <- A AND ((dp+1)(dp)+Y)  N......Z.
			case 0x37:
				LOADDP_Y();
				AND_FROMR1();
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   AND    (X),(Y)     39    1     5   (X) <- (X) AND (Y)         N......Z.
			case 0x39:
				LOADXY();
				*target++ = ANDS_RI(ARM_R0, ARM_R0, ARM_R2, 0, 0);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   AND    dp(d),dp(s) 29    3     6 (dp(d))<-(dp(d)) AND (dp(s)) N......Z.
			case 0x29:
				LOADDPDP();
				*target++ = ANDS_RI(ARM_R0, ARM_R0, ARM_R2, 0, 0);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   AND    dp,#inm     38    3     5  (dp) <- (dp) AND inm        N......Z.
			case 0x38:
				LOADDPIMM();
				*target++ = ANDS_RI(ARM_R0, ARM_R0, ARM_R2, 0, 0);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 5;
				break;

			////////////////////////////////////////////////////////////////////////////
			// Logical OR operations
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   OR     A,#inm      08    2     2     A <- A OR inm            N......Z.
			case 0x08:
				*target++ = ORR_I(APU_REG_A, APU_REG_A, 4, APU_MEM[APU_PC++]);
				*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   OR     A,(X)       06    1     3     A <- A OR (X)            N......Z.
			case 0x06:
				*target++ = MOV_RI(ARM_R1, 0, APU_REG_X, SHIFT_LSR, 24);
				OR_FROMR1();
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   OR     A,dp        04    2     3     A <- A OR (dp)           N......Z.
			case 0x04:
				LOADDP();
				OR_FROMR1();
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   OR     A,dp+X      14    2     4     A <- A OR (dp+X)         N......Z.
			case 0x14:
				LOADDP_OFS(APU_REG_X);
				OR_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   OR     A,labs      05    3     4     A <- A OR (abs)          N......Z.
			case 0x05:
				LOADIMM();
				OR_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   OR     A,labs+X    15    3     5     A <- A OR (abs+X)        N......Z.
			case 0x15:
				LOADIMM_OFS(APU_REG_X);
				OR_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   OR     A,labs+Y    16    3     5     A <- A OR (abs+Y)        N......Z.
			case 0x16:
				LOADIMM_OFS(APU_REG_Y);
				OR_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   OR     A,(dp+X)    07    2     6  A <- A OR ((dp+X+1)(dp+X))  N......Z.
			case 0x07:
				LOADDP_X();
				OR_FROMR1();
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   OR     A,(dp)+Y    17    2     6   A <- A OR ((dp+1)(dp)+Y)   N......Z.
			case 0x17:
				LOADDP_Y();
				OR_FROMR1();
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   OR     (X),(Y)     19    1     5   (X) <- (X) OR (Y)          N......Z.
			case 0x19:
				LOADXY();
				*target++ = ORRS_RI(ARM_R0, ARM_R0, ARM_R2, 0, 0);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   OR     dp(d),dp(s) 09    3     6 (dp(d))<-(dp(d)) OR (dp(s))  N......Z.
			case 0x09:
				LOADDPDP();
				*target++ = ORRS_RI(ARM_R0, ARM_R0, ARM_R2, 0, 0);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   OR		dp,#inm     18    3     5  (dp) <- (dp) OR inm         N......Z.
			case 0x18:
				LOADDPIMM();
				*target++ = ORRS_RI(ARM_R0, ARM_R0, ARM_R2, 0, 0);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 5;
				break;

			////////////////////////////////////////////////////////////////////////////
			// Logical EOR operations
			////////////////////////////////////////////////////////////////////////////
			
			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   EOR    A,#inm      48    2     2     A <- A EOR inm           N......Z.
			case 0x48:
				*target++ = EOR_I(APU_REG_A, APU_REG_A, 4, APU_MEM[APU_PC++]);
				*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   EOR    A,(X)       46    1     3     A <- A EOR (X)           N......Z.
			case 0x46:
				*target++ = MOV_RI(ARM_R1, 0, APU_REG_X, SHIFT_LSR, 24);
				EOR_FROMR1();
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   EOR    A,dp        44    2     3     A <- A EOR (dp)          N......Z.
			case 0x44:
				LOADDP();
				EOR_FROMR1();
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   EOR    A,dp+X      54    2     4     A <- A EOR (dp+X)        N......Z.
			case 0x54:
				LOADDP_OFS(APU_REG_X);
				EOR_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   EOR    A,labs      45    3     4     A <- A EOR (abs)         N......Z.
			case 0x45:
				LOADIMM();
				EOR_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   EOR    A,labs+X    55    3     5     A <- A EOR (abs+X)       N......Z.
			case 0x55:
				LOADIMM_OFS(APU_REG_X);
				EOR_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   EOR    A,labs+Y    56    3     5     A <- A EOR (abs+Y)       N......Z.
			case 0x56:
				LOADIMM_OFS(APU_REG_Y);
				EOR_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   EOR    A,(dp+X)    47    2     6  A <- A EOR ((dp+X+1)(dp+X)) N......Z.
			case 0x47:
				LOADDP_X();
				EOR_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   EOR    A,(dp)+Y    57    2     6   A <- A EOR ((dp+1)(dp)+Y)  N......Z.
			case 0x57:
				LOADDP_Y();
				EOR_FROMR1();
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   EOR    (X),(Y)     59    1     5   (X) <- (X) EOR (Y)         N......Z.
			case 0x59:
				LOADXY();
				*target++ = EORS_RI(ARM_R0, ARM_R0, ARM_R2, 0, 0);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 5;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   EOR    dp(d),dp(s) 49    3     6 (dp(d))<-(dp(d)) EOR (dp(s)) N......Z.
			case 0x49:
				LOADDPDP();
				*target++ = EORS_RI(ARM_R0, ARM_R0, ARM_R2, 0, 0);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 6;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   EOR    dp,#inm     58    3     5  (dp) <- (dp) EOR inm        N......Z.
			case 0x58:
				LOADDPIMM();
				*target++ = EORS_RI(ARM_R0, ARM_R0, ARM_R2, 0, 0);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 5;
				break;

			////////////////////////////////////////////////////////////////////////////
			// Increment commands
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   INC      A         BC    1     2      ++ A                  N......Z.
			case 0xBC:
				*target++ = ADD_I(APU_REG_A, APU_REG_A, 4, 1);
				*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   INC      dp        AB    2     4      ++ (dp)               N......Z.
			case 0xAB:
				LOADDP();
				INC_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   INC     dp+X       BB    2     5      ++ (dp+X)             N......Z.
			case 0xBB:
				LOADDP_OFS(APU_REG_X);
				INC_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   INC     labs       AC    3     5      ++ (abs)              N......Z.
			case 0xAC:
				LOADIMM();
				INC_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   INC      X         3D    1     2      ++ X                  N......Z.
			case 0x3D:
				*target++ = ADD_I(APU_REG_X, APU_REG_X, 4, 1);
				*target++ = MOVS_RI(APU_REG_X, 0, APU_REG_X, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   INC      Y         FC    1     2      ++ Y                  N......Z.
			case 0xFC:
				*target++ = ADD_I(APU_REG_Y, APU_REG_Y, 4, 1);
				*target++ = MOVS_RI(APU_REG_Y, 0, APU_REG_Y, 0, 0);
				cycles += 2;
				break;

			////////////////////////////////////////////////////////////////////////////
			// Decrement commands
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   DEC      A         9C    1     2      -- A                  N......Z.
			case 0x9C:
				*target++ = SUB_I(APU_REG_A, APU_REG_A, 4, 1);
				*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   DEC      dp        8B    2     4      -- (dp)               N......Z.
			case 0x8B:
				LOADDP();
				DEC_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   DEC     dp+X       9B    2     5      -- (dp+X)             N......Z.
			case 0x9B:
				LOADDP_OFS(APU_REG_X);
				DEC_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   DEC     labs       8C    3     5      -- (abs)              N......Z.
			case 0x8C:
				LOADIMM();
				DEC_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   DEC      X         1D    1     2      -- X                  N......Z.
			case 0x1D:
				*target++ = SUB_I(APU_REG_X, APU_REG_X, 4, 1);
				*target++ = MOVS_RI(APU_REG_X, 0, APU_REG_X, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   DEC      Y         DC    1     2      -- Y                  N......Z.
			case 0xDC:
				*target++ = SUB_I(APU_REG_Y, APU_REG_Y, 4, 1);
				*target++ = MOVS_RI(APU_REG_Y, 0, APU_REG_Y, 0, 0);
				cycles += 2;
				break;

			////////////////////////////////////////////////////////////////////////////
			// Shift and rotation commands
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   ASL      A         1C    1     2      C << A      <<0       N......ZC
			case 0x1C:
				*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, SHIFT_LSL, 1);
				cycles += 2;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   ASL      dp        0B    2     4      C << (dp)   <<0       N......ZC
			case 0x0B:
				LOADDP();
				ASL_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   ASL     dp+X       1B    2     5      C << (dp+X) <<0       N......ZC
			case 0x1B:
				LOADDP_OFS(APU_REG_X);
				ASL_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   ASL     labs       CC    3     5      C << (abs)  <<0       N......ZC
			case 0xCC:
				LOADIMM();
				ASL_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   LSR      A         5C    1     2      0 >> A      <<C       N......ZC
			case 0x5C:
				*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, SHIFT_LSR, 25);
				*target++ = MOV_RI(APU_REG_A, 0, APU_REG_A, SHIFT_LSL, 24);
				cycles += 2;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   LSR      dp        4B    2     4      0 >> (dp)   <<C       N......ZC
			case 0x4B:
				LOADDP();
				LSR_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   LSR     dp+X       5B    2     5      0 >> (dp+X) <<C       N......ZC
			case 0x5B:
				LOADDP_OFS(APU_REG_X);
				LSR_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   LSR     labs       4C    3     5      0 >> (abs)  <<C       N......ZC
			case 0x4C:
				LOADIMM();
				LSR_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   ROL      A         3C    1     2      C << A      <<C       N......ZC
			case 0x3C:
				*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, SHIFT_LSL, 1);
				*target++ = ORRCS_I(APU_REG_A, APU_REG_A, 4, 1);
				*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   ROL      dp        2B    2     4      C << (dp)   <<C       N......ZC
			case 0x2B:
				LOADDP();
				ROL_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   ROL     dp+X       3B    2     5      C << (dp+X) <<C       N......ZC
			case 0x3B:
				LOADDP_OFS(APU_REG_X);
				ROL_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   ROL     labs       2C    3     5      C << (abs)  <<C       N......ZC
			case 0x2C:
				LOADIMM();
				ROL_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   ROR      A         7C    1     2      C >> A      <<C       N......ZC
			case 0x7C:
				*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, SHIFT_LSR, 25);
				*target++ = ORRCS_I(APU_REG_A, APU_REG_A, 0, 0x80);
				*target++ = MOV_RI(APU_REG_A, 0, APU_REG_A, SHIFT_LSL, 24);
				*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   ROR      dp        6B    2     4      C >> (dp)   <<C       N......ZC
			case 0x6B:
				LOADDP();
				ROR_FROMR1();
				cycles += 4;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   ROR     dp+X       7B    2     5      C >> (dp+X) <<C       N......ZC
			case 0x7B:
				LOADDP_OFS(APU_REG_X);
				ROR_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   ROR     labs       6C    3     5      C >> (abs)  <<C       N......ZC
			case 0x6C:
				LOADIMM();
				ROR_FROMR1();
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   XCN      A         9F    1     5      A(7-4) <-> A(3-0)     N......Z.
			case 0x9F:
				*target++ = MOV_RI(ARM_R0, 0, APU_REG_A, SHIFT_LSR, 28);
				*target++ = MOV_RI(APU_REG_A, 0, APU_REG_A, SHIFT_LSL, 4);
				*target++ = ORR_RI(APU_REG_A, APU_REG_A, ARM_R0, SHIFT_LSL, 24);
				*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, 0, 0);
				cycles += 5;
				break;

			////////////////////////////////////////////////////////////////////////////
			// 16 bit transmission commands
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   MOVW     YA,dp     BA    2     5     YA  -> (dp+1)(dp)      N......Z.
			case 0xBA:
				LOADDP();
				*target++ = LDRB_R(APU_RAMBASE, APU_REG_A, ARM_R1);
				EmitMemReadCheckR1(APU_REG_A);
				*target++ = MOV_RI(APU_REG_A, 0, APU_REG_A, SHIFT_LSL, 24);
				*target++ = ADD_I(ARM_R1, ARM_R1, 0, 1);
				*target++ = LDRB_R(APU_RAMBASE, APU_REG_Y, ARM_R1);
				EmitMemReadCheckR1(APU_REG_Y);
				*target++ = MOV_RI(APU_REG_Y, 0, APU_REG_Y, SHIFT_LSL, 24);
				*target++ = MOV_RI(ARM_R0, 0, APU_REG_A, SHIFT_LSR, 8);
				*target++ = ORRS_RI(ARM_R0, ARM_R0, APU_REG_Y, 0, 0);
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   MOVW     dp,YA     DA    2     4    (dp+1)(dp) <- YA        .........
			case 0xDA:
				LOADDP();
				*target++ = MOV_RI(ARM_R0, 0, APU_REG_A, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				*target++ = ADD_I(ARM_R1, ARM_R1, 0, 1);
				*target++ = MOV_RI(ARM_R0, 0, APU_REG_Y, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 4;
				break;

			////////////////////////////////////////////////////////////////////////////
			// 16 bit operation commands
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   INCW     dp        3A    2     6  Increment dp memory pair  N......Z.
			case 0x3A:
				LOADDP();
				// Load the (dp) memory pair
				*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); 
				*target++ = ADD_I(ARM_R3, ARM_R1, 0, 1);
				*target++ = LDRB_R(APU_RAMBASE, ARM_R2, ARM_R3);
				// r2 is high 8, r0 is low 8
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSL, 16);
				*target++ = ORR_RI(ARM_R2, ARM_R0, ARM_R2, SHIFT_LSL, 24);
				// Increment and flag test
				*target++ = ADD_I(ARM_R2, ARM_R2, 8, 1);
				*target++ = MOVS_RI(ARM_R2, 0, ARM_R2, 0, 0);
				// Store it back at (dp)
				*target++ = MOV_RI(ARM_R2, 0, ARM_R2, SHIFT_LSR, 16);
				*target++ = STRB_R(APU_RAMBASE, ARM_R2, ARM_R1);
				EmitMemWriteCheckR1Reg(ARM_R2);
				*target++ = MOV_RI(ARM_R2, 0, ARM_R2, SHIFT_LSR, 8);
				*target++ = ADD_I(ARM_R1, ARM_R1, 0, 1);
				*target++ = STRB_R(APU_RAMBASE, ARM_R2, ARM_R1);
				EmitMemWriteCheckR1Reg(ARM_R2);
				cycles += 6;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   DECW     dp        1A    2     6  Decrement dp memory pair  N......Z.
			case 0x1A:
				LOADDP();
				// Load the (dp) memory pair
				*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); 
				*target++ = ADD_I(ARM_R3, ARM_R1, 0, 1);
				*target++ = LDRB_R(APU_RAMBASE, ARM_R2, ARM_R3);
				// r2 is high 8, r0 is low 8
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSL, 16);
				*target++ = ORR_RI(ARM_R2, ARM_R0, ARM_R2, SHIFT_LSL, 24);
				// Decrement and flag test
				*target++ = SUB_I(ARM_R2, ARM_R2, 8, 1);
				*target++ = MOVS_RI(ARM_R2, 0, ARM_R2, 0, 0);
				// Store it back at (dp)
				*target++ = MOV_RI(ARM_R2, 0, ARM_R2, SHIFT_LSR, 16);
				*target++ = STRB_R(APU_RAMBASE, ARM_R2, ARM_R1);
				EmitMemWriteCheckR1Reg(ARM_R2);
				*target++ = MOV_RI(ARM_R2, 0, ARM_R2, SHIFT_LSR, 8);
				*target++ = ADD_I(ARM_R1, ARM_R1, 0, 1);
				*target++ = STRB_R(APU_RAMBASE, ARM_R2, ARM_R1);
				EmitMemWriteCheckR1Reg(ARM_R2);
				cycles += 6;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   ADDW    YA,dp      7A    2     5   YA  <- YA + (dp+1)(dp)   NV..H..ZC
			case 0x7A:
				LOADDP();
				// Load the (dp) memory pair
				*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); 
				EmitMemReadCheckR1(ARM_R0);
				*target++ = ADD_I(ARM_R1, ARM_R1, 0, 1);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSL, 16);
				*target++ = LDRB_R(APU_RAMBASE, ARM_R2, ARM_R1);
				EmitMemReadCheckR1(ARM_R2);
				*target++ = ORR_RI(ARM_R0, ARM_R0, ARM_R2, SHIFT_LSL, 24);
				// Build YA and add R0 to it
				*target++ = ORR_RI(APU_REG_A, APU_REG_Y, APU_REG_A, SHIFT_LSR, 8);
				*target++ = ADDS_RI(APU_REG_A, APU_REG_A, ARM_R0, 0, 0);
				*target++ = AND_I(APU_REG_Y, APU_REG_A, 4, 0xff);
				*target++ = MOV_RI(APU_REG_A, 0, APU_REG_A, SHIFT_LSL, 8);
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   SUBW    YA,dp      9A    2     5   YA  <- YA - (dp+1)(dp)   NV..H..ZC
			case 0x9A:
				LOADDP();
				// Load the (dp) memory pair
				*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); 
				EmitMemReadCheckR1(ARM_R0);
				*target++ = ADD_I(ARM_R1, ARM_R1, 0, 1);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSL, 16);
				*target++ = LDRB_R(APU_RAMBASE, ARM_R2, ARM_R1);
				EmitMemReadCheckR1(ARM_R2);
				*target++ = ORR_RI(ARM_R0, ARM_R0, ARM_R2, SHIFT_LSL, 24);
				// Build YA and sub R0 to it
				*target++ = ORR_RI(APU_REG_A, APU_REG_Y, APU_REG_A, SHIFT_LSR, 8);
				*target++ = SUBS_RI(APU_REG_A, APU_REG_A, ARM_R0, 0, 0);
				*target++ = AND_I(APU_REG_Y, APU_REG_A, 4, 0xff);
				*target++ = MOV_RI(APU_REG_A, 0, APU_REG_A, SHIFT_LSL, 8);
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   CMPW    YA,dp      5A    2     4     YA - (dp+1)(dp)        N......Z.
			case 0x5A:
				LOADDP();
				// Load the (dp) memory pair
				*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1); 
				EmitMemReadCheckR1(ARM_R0);
				*target++ = ADD_I(ARM_R1, ARM_R1, 0, 1);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSL, 16);
				*target++ = LDRB_R(APU_RAMBASE, ARM_R2, ARM_R1);
				EmitMemReadCheckR1(ARM_R2);
				*target++ = ORR_RI(ARM_R0, ARM_R0, ARM_R2, SHIFT_LSL, 24);
				// Build YA and sub R0 from it
				*target++ = ORR_RI(ARM_R1, APU_REG_Y, APU_REG_A, SHIFT_LSR, 8);
				*target++ = SUB_RI(ARM_R1, ARM_R1, ARM_R0, 0, 0);
				*target++ = MOVS_RI(ARM_R1, 0, ARM_R1, 0, 0);
				cycles += 4;
				break;

			////////////////////////////////////////////////////////////////////////////
			// Multiplication and division commands
			////////////////////////////////////////////////////////////////////////////
			
			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   MUL      YA        CF    1     9    YA(16 bits) <- Y * A    N......Z.
			case 0xCF:
				*target++ = MOV_RI(ARM_R0, 0, APU_REG_Y, SHIFT_LSR, 16);
				*target++ = MOV_RI(ARM_R1, 0, APU_REG_A, SHIFT_LSR, 16);
				*target++ = MUL(ARM_R0, ARM_R0, ARM_R1);
				*target++ = MOVS_RI(ARM_R0, 0, ARM_R0, 0, 0);
				*target++ = MOV_RI(APU_REG_A, 0, ARM_R0, SHIFT_LSL, 8);
				*target++ = MOV_RI(ARM_R0, 0, ARM_R0, SHIFT_LSR, 24);
				*target++ = MOV_RI(APU_REG_Y, 0, ARM_R0, SHIFT_LSL, 24);
				cycles += 9;
				break;
			
			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   DIV      YA,X      9E    1    12    Q:A B:Y <- YA / X       NV..H..Z.
			case 0x9E:
				*target++ = CMPS_I(0, APU_REG_X, 0, 0);
				// Branch to zero handling
				*target++ = BEQ(9);

				// Do the actual unsigned divide, clear overflow as well
				*target++ = MOV_RI(ARM_R0, 0, APU_REG_Y, SHIFT_LSR, 16);
				*target++ = ORR_RI(ARM_R0, ARM_R0, APU_REG_A, SHIFT_LSR, 24);
				*target++ = MOV_RI(ARM_R1, 0, APU_REG_X, SHIFT_LSR, 24);
				// Divide takes r0 number, r1 denom
				// returns r0 = number / denom, r1 = number % denom, r2 = abs(r0)
				*target++ = SWI(0x06 << 16);
				*target++ = MOV_RI(APU_REG_A, 0, ARM_R0, SHIFT_LSL, 24);
				*target++ = MOV_RI(APU_REG_Y, 0, ARM_R1, SHIFT_LSL, 24);

				// Clear V flag, bit 28
				*target++ = MRS(ARM_R2);
				*target++ = BIC_I(ARM_R2, ARM_R2, 4, 0x10);
				*target++ = MSR_R(ARM_R2);

				// Jump to flag update
				*target++ = B(4);

				//
				// Zero divisor, set overflow
				//
				*target++ = MOV_I(APU_REG_Y, 0, 4, 0xFF);
				*target++ = MOV_I(APU_REG_A, 0, 4, 0xFF);
				// Set V flag, bit 28
				*target++ = MRS(ARM_R2);
				*target++ = ORR_I(ARM_R2, ARM_R2, 4, 0x10);
				*target++ = MSR_R(ARM_R2);

				// Set the N and Z flags
				*target++ = MOVS_RI(APU_REG_A, 0, APU_REG_A, 0, 0);
				cycles += 12;
				break;

			////////////////////////////////////////////////////////////////////////////
			// Decimal compensation commands
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   DAA       A        DF    1     3    decimal adjust for add  N......ZC
			case 0xDF:
				unimpl(0xDF, startPC);
				cycles += 3;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   DAS       A        BE    1     3    decimal adjust for sub  N......ZC
			case 0xBE:
				unimpl(0xBE, startPC);
				cycles += 3;
				break;
			
			////////////////////////////////////////////////////////////////////////////
			// Branching commands
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   BRA     rel        2F    2     4   branch always                  ...
			case 0x2F:
				cycles += 4;
				// Put the PC where it is supposed to be
				val = ((s8*)APU_MEM)[APU_PC++];
				// Update stored PC to be consistent
				*target++ = ADD_I(APU_REG_PC, APU_REG_PC, 0, APU_PC - startPC);
				if (val < 0) *target++ = SUB_I(APU_REG_PC, APU_REG_PC, 0, -val);
				else *target++ = ADD_I(APU_REG_PC, APU_REG_PC, 0, val);

				// Store the flags
				*target++ = MRS(ARM_R4);
				// Record the number of cycles taken here
				*target++ = MOV_I(ARM_R3, 0, 0, cycles);
				// Branch back to the dispatch loop
				*target++ = MOV_RI(ARM_PC, 0, ARM_LR, 0, 0);
				goto doneRecomp;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   BEQ     rel        F0    2    2/4  branch on Z=1                  ...
			case 0xF0:
				cycles += 4;
				*target++ = MRS(ARM_R4);
				BRANCH_REL(EQ_RI);
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   BNE     rel        D0    2    2/4  branch on Z=0                  ...
			case 0xD0:
				cycles += 4;
				*target++ = MRS(ARM_R4);
				BRANCH_REL(NE_RI);
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   BCS     rel        B0    2    2/4  branch on C=1                  ... 
			case 0xB0:
				cycles += 4;
				*target++ = MRS(ARM_R4);
				BRANCH_REL(CS_RI);
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   BCC     rel        90    2    2/4  branch on C=0                  ...
			case 0x90:
				cycles += 4;
				*target++ = MRS(ARM_R4);
				BRANCH_REL(CC_RI);
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   BVS     rel        70    2    2/4  branch on V=1                  ...
			case 0x70:
				cycles += 4;
				*target++ = MRS(ARM_R4);
				BRANCH_REL(VS_RI);
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   BVC     rel        50    2    2/4  branch on V=0                  ...
			case 0x50:
				cycles += 4;
				*target++ = MRS(ARM_R4);
				BRANCH_REL(VC_RI);
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   BMI     rel        30    2    2/4  branch on N=1                  ...
			case 0x30:
				cycles += 4;
				*target++ = MRS(ARM_R4);
				BRANCH_REL(MI_RI);
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   BPL     rel        10    2    2/4  branch on N=0                  ...
			case 0x10:
				cycles += 4;
				*target++ = MRS(ARM_R4);
				BRANCH_REL(PL_RI);
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   BBS   dp.bit,rel   x3    3    5/7  branch on dp.bit=1             ...
			case 0x03: BRANCH_DPBIT_SET(0); break;
			case 0x23: BRANCH_DPBIT_SET(1); break;
			case 0x43: BRANCH_DPBIT_SET(2); break;
			case 0x63: BRANCH_DPBIT_SET(3); break;
			case 0x83: BRANCH_DPBIT_SET(4); break;
			case 0xA3: BRANCH_DPBIT_SET(5); break;
			case 0xC3: BRANCH_DPBIT_SET(6); break;
			case 0xE3: BRANCH_DPBIT_SET(7); break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   BBC   dp.bit,rel   y3    3    5/7  branch on dp.bit=0             ...
			case 0x13: BRANCH_DPBIT_CLR(0); break;
			case 0x33: BRANCH_DPBIT_CLR(1); break;
			case 0x53: BRANCH_DPBIT_CLR(2); break;
			case 0x73: BRANCH_DPBIT_CLR(3); break;
			case 0x93: BRANCH_DPBIT_CLR(4); break;
			case 0xB3: BRANCH_DPBIT_CLR(5); break;
			case 0xD3: BRANCH_DPBIT_CLR(6); break;
			case 0xF3: BRANCH_DPBIT_CLR(7); break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   CBNE   dp,rel      2E    3    5/7  compare A with (dp) then BNE   ...
			case 0x2E:
				cycles += 7;
				LOADDP();
				*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemReadCheckR1(ARM_R0);
				// Store the flag before we do the test, so we don't modify it
				*target++ = MRS(ARM_R4);
				// Compare A with (dp)
 				*target++ = CMPS_RI(0, APU_REG_A, ARM_R0, SHIFT_LSL, 24);
				BRANCH_REL(NE_RI);				
				goto doneRecomp;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   CBNE  dp+X,rel     DE    3    6/8  compare A with (dp+X) then BNE ...
			case 0xDE:
				cycles += 8;
				LOADDP_OFS(APU_REG_X);
				*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemReadCheckR1(ARM_R0);
				// Store the flag before we do the test, so we don't modify it
				*target++ = MRS(ARM_R4);
				// Compare A with (dp)
				*target++ = CMPS_RI(0, APU_REG_A, ARM_R0, SHIFT_LSL, 24);
				BRANCH_REL(NE_RI);
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   DBNZ   dp,rel      6E    3    5/7  decrement memory (dp) then JNZ ...
			case 0x6E:
				cycles += 7;
				LOADDP();
				*target++ = LDRSB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemReadCheckR1(ARM_R0);
				// Store the flag before we do the test, so we don't modify it
				*target++ = MRS(ARM_R4);
				// Decrement (dp), then store back
				*target++ = SUBS_I(ARM_R0, ARM_R0, 0, 1);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1Reg(ARM_R0);
				BRANCH_REL(NE_RI);
				goto doneRecomp;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   DBNZ    Y,rel      FE    2    4/6  decrement Y then JNZ           ...
			case 0xFE:
				cycles += 6;
				// Store the flag before we do the test, so we don't modify it
				*target++ = MRS(ARM_R4);
				// Decrement Y and test for NZ
				*target++ = SUBS_I(APU_REG_Y, APU_REG_Y, 4, 1);
				BRANCH_REL(NE_RI);
				goto doneRecomp;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   JMP     labs       5F    3     3   jump to new location           ...
			case 0x5F:
				cycles += 3;

				*target++ = MRS(ARM_R4);

				// Set PC to called location (equivalent to LOADIMM)
				*target++ = MOV_I(APU_REG_PC, 0, 0, APU_MEM[APU_PC++]);
				*target++ = ORR_I(APU_REG_PC, APU_REG_PC, 12, APU_MEM[APU_PC++]);

				// Record the number of cycles taken here
				*target++ = MOV_I(ARM_R3, 0, 0, cycles);
				// Branch back to the dispatch loop
				*target++ = MOV_RI(ARM_PC, 0, ARM_LR, 0, 0);

				goto doneRecomp;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   JMP    (labs+X)    1F    3     6   PC <- (abs+X+1)(abs+X)         ...
			case 0x1F:
				cycles += 6;

				*target++ = MRS(ARM_R4);

				// Set PC to called location (equivalent to LOADIMM)
				LOADIMM_OFS(APU_REG_X);
				*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				*target++ = ADD_I(ARM_R1, ARM_R1, 0, 1);
				*target++ = LDRB_R(APU_RAMBASE, ARM_R2, ARM_R1);
				*target++ = ORR_RI(APU_REG_PC, ARM_R0, ARM_R2, SHIFT_LSL, 8);

				// Record the number of cycles taken here
				*target++ = MOV_I(ARM_R3, 0, 0, cycles);
				// Branch back to the dispatch loop
				*target++ = MOV_RI(ARM_PC, 0, ARM_LR, 0, 0);

				goto doneRecomp;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                 Flag
			//   CALL     labs      3F    3     8   subroutine call          ........ 
			case 0x3F:
				cycles += 8;

				// Store the flags
				*target++ = MRS(ARM_R4);

				// Load the address to jump to into r1
				LOADIMM();

				// Update stored PC to be consistent
				*target++ = ADD_I(APU_REG_PC, APU_REG_PC, 0, APU_PC - startPC);
				// Push on the APU_PC to return to
				PUSH_WORD(APU_REG_PC);

				// Set PC to called location
				*target++ = MOV_RI(APU_REG_PC, 0, ARM_R1, 0, 0);
				// Record the number of cycles taken here
				*target++ = MOV_I(ARM_R3, 0, 0, cycles);
				// Branch back to the dispatch loop
				*target++ = MOV_RI(ARM_PC, 0, ARM_LR, 0, 0);
				goto doneRecomp;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                 Flag
			//   PCALL   upage      4F    2     6   upage call               ........ 
			case 0x4F:
				APU_PC++;
				unimpl(0x4F, startPC);
				cycles += 6;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                 Flag
			//   TCALL     n        n1    1     8   table call               ........ 
			case 0x01: TCALL(0); break;
			case 0x11: TCALL(1); break;
			case 0x21: TCALL(2); break;
			case 0x31: TCALL(3); break;
			case 0x41: TCALL(4); break;
			case 0x51: TCALL(5); break;
			case 0x61: TCALL(6); break;
			case 0x71: TCALL(7); break;
			case 0x81: TCALL(8); break;
			case 0x91: TCALL(9); break;
			case 0xA1: TCALL(10); break;
			case 0xB1: TCALL(11); break;
			case 0xC1: TCALL(12); break;
			case 0xD1: TCALL(13); break;
			case 0xE1: TCALL(14); break;
			case 0xF1: TCALL(15); break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                 Flag
			//   BRK                0F    1     8   software interrupt       ...1.0..                                                                     
			case 0x0F:
				unimpl(0x0F, startPC);
				cycles += 8;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                 Flag
			//   RET                6F    1     5   return from subroutine   ........ 
			case 0x6F:
				cycles += 5;

				// Load the address to jump to into r1
				*target++ = MRS(ARM_R4);

				// This is the location we should return to
				POP_WORD(APU_REG_PC);

				// Record the number of cycles taken here
				*target++ = MOV_I(ARM_R3, 0, 0, cycles);
				// Branch back to the dispatch loop
				*target++ = MOV_RI(ARM_PC, 0, ARM_LR, 0, 0);

				goto doneRecomp;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                 Flag
			//   RET1               7F    1     6   return from interrupt   (Restored)
			case 0x7F:
				unimpl(0x7F, startPC);
				cycles += 5;
				break;

			////////////////////////////////////////////////////////////////////////////
			// Stack operation commands
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   PUSH      A        2D    1     4    push A to stack         .........
			case 0x2D:
				// put the value at SP, then decrement the SP
				*target++ = MOV_RI(ARM_R0, 0, APU_REG_A, SHIFT_LSR, 24);
				PUSH_BYTE(ARM_R0);
				cycles += 4;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   PUSH      X        4D    1     4    push X to stack         .........
			case 0x4D:
				*target++ = MOV_RI(ARM_R0, 0, APU_REG_X, SHIFT_LSR, 24);
				PUSH_BYTE(ARM_R0);
				cycles += 4;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   PUSH      Y        6D    1     4    push Y to stack         .........
			case 0x6D:
				*target++ = MOV_RI(ARM_R0, 0, APU_REG_Y, SHIFT_LSR, 24);
				PUSH_BYTE(ARM_R0);
				cycles += 4;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   PUSH     PSW       0D    1     4    push PSW to stack       .........
			case 0x0D:
				PACK_PSW();
				PUSH_BYTE(ARM_R0);
				cycles += 4;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   POP       A        AE    1     4    pop A from stack        .........
			case 0xAE:
				POP_BYTE(APU_REG_A);
				*target++ = MOV_RI(APU_REG_A, 0, APU_REG_A, SHIFT_LSL, 24);
				cycles += 4;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   POP       X        CE    1     4    pop X from stack        .........
			case 0xCE:
				POP_BYTE(APU_REG_X);
				*target++ = MOV_RI(APU_REG_X, 0, APU_REG_X, SHIFT_LSL, 24);
				cycles += 4;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   POP       Y        EE    1     4    pop Y from stack        .........
			case 0xEE:
				POP_BYTE(APU_REG_Y);
				*target++ = MOV_RI(APU_REG_Y, 0, APU_REG_Y, SHIFT_LSL, 24);
				cycles += 4;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   POP      PSW       8E    1     4    pop PSW from stack     (Restored)
			case 0x8E:
				POP_BYTE(ARM_R0);
				UNPACK_PSW();
				cycles += 4;
				break;

			////////////////////////////////////////////////////////////////////////////
			// Bit operation commands
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   SET1    dip.bit    x2    2     4  set direct page bit       .........
			case 0x02: SET_DP_BIT(0); break;
			case 0x22: SET_DP_BIT(1); break;
			case 0x42: SET_DP_BIT(2); break;
			case 0x62: SET_DP_BIT(3); break;
			case 0x82: SET_DP_BIT(4); break;
			case 0xA2: SET_DP_BIT(5); break;
			case 0xC2: SET_DP_BIT(6); break;
			case 0xE2: SET_DP_BIT(7); break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   CLR1    dip.bit    y2    2     4  clear direct page bit     .........
			case 0x12: CLEAR_DP_BIT(0); break;
			case 0x32: CLEAR_DP_BIT(1); break;
			case 0x52: CLEAR_DP_BIT(2); break;
			case 0x72: CLEAR_DP_BIT(3); break;
			case 0x92: CLEAR_DP_BIT(4); break;
			case 0xB2: CLEAR_DP_BIT(5); break;
			case 0xD2: CLEAR_DP_BIT(6); break;
			case 0xF2: CLEAR_DP_BIT(7); break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   TSET1    labs      0E    3     6 test and set bits with A   N......Z.
			case 0x0E:
				LOADIMM();
				*target++ = LDRB_R(APU_RAMBASE, ARM_R2, ARM_R1);
				*target++ = ORR_RI(ARM_R0, ARM_R2, APU_REG_A, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				*target++ = AND_RI(ARM_R2, APU_REG_A, ARM_R2, SHIFT_LSL, 24);
				*target++ = MOVS_RI(ARM_R2, 0, ARM_R2, 0, 0);
				EmitMemWriteCheckR1();
				cycles += 6;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   TCLR1    labs      4E    3     6 test and clear bits with A N......Z.
			case 0x4E:
				LOADIMM();
				*target++ = LDRB_R(APU_RAMBASE, ARM_R2, ARM_R1);
				*target++ = BIC_RI(ARM_R0, ARM_R2, APU_REG_A, SHIFT_LSR, 24);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				*target++ = AND_RI(ARM_R2, APU_REG_A, ARM_R2, SHIFT_LSL, 24);
				*target++ = MOVS_RI(ARM_R2, 0, ARM_R2, 0, 0);
				EmitMemWriteCheckR1();
				cycles += 6;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   AND1   C,mem.bit   4A    3     4  C <- C AND (mem.bit)      ........C
			case 0x4A:
				LOADIMM();
				// r1 is address, high 13 bits are actual address, low 3 bits are bit in byte
				*target++ = AND_I(ARM_R2, ARM_R1, 0, 0x7);
				*target++ = MOV_RI(ARM_R1, 0, ARM_R1, SHIFT_LSR, 3);
				*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemReadCheckR1(ARM_R0);
				*target++ = MRS(ARM_R1);
				*target++ = MOVCC_I(ARM_R3, 0, 0, 0);
				*target++ = MOVCS_I(ARM_R3, 0, 0, 1);
				*target++ = AND_R(ARM_R3, ARM_R3, ARM_R0, SHIFT_LSR, ARM_R2);
				// Carry flag is bit 29
				*target++ = BIC_I(ARM_R1, ARM_R1, 4, 0x20);
				*target++ = ORR_RI(ARM_R1, ARM_R1, ARM_R3, SHIFT_LSL, ARM_FLAG_C);
				*target++ = MSR_R(ARM_R1);
				cycles += 4;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   AND1   C,/mem.bit  6A    3     4  C <- C AND !(mem.bit)     ........C
			case 0x6A:
				LOADIMM();
				unimpl(0x6A, startPC);
				cycles += 4;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   OR1    C,mem.bit   0A    3     5  C <- C OR  (mem.bit)      ........C
			case 0x0A:
				LOADIMM();
				unimpl(0x0A, startPC);
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   OR1    C,/mem.bit  2A    3     5  C <- C OR  !(mem.bit)     ........C
			case 0x2A:
				LOADIMM();
				unimpl(0x2A, startPC);
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   EOR1   C,mem.bit   8A    3     5  C <- C EOR (mem.bit)      ........C
			case 0x8A:
				LOADIMM();
				unimpl(0x8A, startPC);
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   NOT1   mem.bit     EA    3     5  complement (mem.bit)      .........
			case 0xEA:
				LOADIMM();
				// r1 is address, high 13 bits are actual address, low 3 bits are bit in byte
				*target++ = AND_I(ARM_R2, ARM_R1, 0, 0x7);
				*target++ = MOV_RI(ARM_R1, 0, ARM_R1, SHIFT_LSR, 3);
				*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				*target++ = MOV_I(ARM_R3, 0, 0, 1);
				*target++ = EOR_R(ARM_R0, ARM_R0, ARM_R3, SHIFT_LSL, ARM_R2);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 5;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   MOV1   C,mem.bit   AA    3     4  C <- (mem.bit)            ........C
			case 0xAA:
				LOADIMM();
				// r1 is address, high 13 bits are actual address, low 3 bits are bit in byte
				*target++ = AND_I(ARM_R2, ARM_R1, 0, 0x7);
				*target++ = MOV_RI(ARM_R1, 0, ARM_R1, SHIFT_LSR, 3);
				*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemReadCheckR1(ARM_R0);
				*target++ = MRS(ARM_R1);
				*target++ = MOV_I(ARM_R3, 0, 0, 1);
				*target++ = AND_R(ARM_R3, ARM_R3, ARM_R0, SHIFT_LSR, ARM_R2);
				// Carry flag is bit 29
				*target++ = BIC_I(ARM_R1, ARM_R1, 4, 0x20);
				*target++ = ORR_RI(ARM_R1, ARM_R1, ARM_R3, SHIFT_LSL, ARM_FLAG_C);
				*target++ = MSR_R(ARM_R1);
				cycles += 4;
				break;

			// Mnemonic  Operand   Code Bytes Cycles Operation                  Flag
			//   MOV1   mem.bit,C   CA    3     6  C -> (mem.bit)            .........
			case 0xCA:
				LOADIMM();
				*target++ = AND_I(ARM_R2, ARM_R1, 0, 0x7);
				*target++ = MOV_RI(ARM_R1, 0, ARM_R1, SHIFT_LSR, 3);
				*target++ = LDRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				*target++ = MOV_I(ARM_R3, 0, 0, 1);
				*target++ = BICCC_R(ARM_R0, ARM_R0, ARM_R3, SHIFT_LSL, ARM_R2);
				*target++ = ORRCS_R(ARM_R0, ARM_R0, ARM_R3, SHIFT_LSL, ARM_R2);
				*target++ = STRB_R(APU_RAMBASE, ARM_R0, ARM_R1);
				EmitMemWriteCheckR1();
				cycles += 6;
				break;

			////////////////////////////////////////////////////////////////////////////
			// Program status flag commands
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CLRC               60    1     2   clear carry flag          .......0 
			case 0x60:
				*target++ = MRS(ARM_R0);
				// Carry flag is bit 29, so to get to it, we ROR 0x2 by 2(x2), to get it
				*target++ = BIC_I(ARM_R0, ARM_R0, 2, 0x2);
				*target++ = MSR_R(ARM_R0);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   SETC               80    1     2   set carry flag            .......1 
			case 0x80:
				*target++ = MRS(ARM_R0);
				// Carry flag is bit 29, so to get to it, we ROR 0x2 by 2(x2), to get it
				*target++ = ORR_I(ARM_R0, ARM_R0, 2, 0x2);
				*target++ = MSR_R(ARM_R0);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   NOTC               ED    1     3   complement carry flag     .......C 
			case 0xED:
				*target++ = MRS(ARM_R0);
				// Carry flag is bit 29, so to get to it, we ROR 0x2 by 2(x2), to get it
				*target++ = EOR_I(ARM_R0, ARM_R0, 2, 0x2);
				*target++ = MSR_R(ARM_R0);
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CLRV               E0    1     2   clear V and H             .0..0... 
			case 0xE0:
				*target++ = MRS(ARM_R0);
				// TODO: handle clearing of H bit
				// V is bit 28, H is ignored
				*target++ = BIC_I(ARM_R0, ARM_R0, 2, 0x1);
				*target++ = MSR_R(ARM_R0);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   CLRP               20    1     2   clear direct page flag    ..0..... 
			case 0x20:
				*target++ = MOV_I(APU_DP, 0, 0, 0);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   SETP               40    1     2   set direct page flag      ..1..0..
			case 0x40:
				// 0x100 = 0x1 ROR 24
				*target++ = MOV_I(APU_DP, 0, 12, 0x1);
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   EI                 A0    1     3  set interrup enable flag   .....1..
			case 0xA0:
//				unimpl(0xA0, startPC);
// TODO: do a somewhat expensive operation here to set flags
//				*target++ = ORR_I(APU_FLAGS, APU_FLAGS, 0, 1);
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   DI                 C0    1     3  clear interrup enable flag .....0..
			case 0xC0:
//				unimpl(0xC0, startPC);
// TODO: do a somewhat expensive operation here to set flags
//				*target++ = BIC_I(APU_FLAGS, APU_FLAGS, 0, 1);
				cycles += 3;
				break;

			////////////////////////////////////////////////////////////////////////////
			// Other commands
			////////////////////////////////////////////////////////////////////////////

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   NOP               00    1     2    no operation               .........
			case 0x00:
				cycles += 2;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   SLEEP             EF    1     3    standby SLEEP mode         .........
			case 0xEF:
				unimpl(0xEF, startPC);
				cycles += 3;
				break;

			// Mnemonic  Operand  Code Bytes Cycles Operation                    Flag
			//   STOP              FF    1     3    standby STOP mode          .........
			case 0xFF:
				unimpl(0xFF, startPC);
				cycles += 3;
				break;

			default:
				unimpl(APU_MEM[APU_PC-1], startPC);
				break;
		}
/*		// DEBUGGING CRAP -- single step code
		*target++ = ADD_I(APU_REG_PC, APU_REG_PC, 0, APU_PC - startPC);
		// Store the flags
		*target++ = MRS(ARM_R4);
		// Record the number of cycles taken here
		*target++ = MOV_I(ARM_R3, 0, 0, cycles);
		// Branch back to the dispatch loop
		*target++ = MOV_RI(ARM_PC, 0, ARM_LR, 0, 0);
		break;*/
	}
doneRecomp:
	return target - startTarget;
}
