	.SECTION .iwram, "ax", %progbits
	.ARM
	.ALIGN

.GLOBAL		CpuFastSet
CpuFastSet:
	swi (12<<16)
	bx lr
	
@ r0 - Address of stored state
.GLOBAL		SetupAndEnterPageLookupAndBranch
SetupAndEnterPageLookupAndBranch:
	stmfd sp!, {r4-r12,lr}
	stmfd sp!, {r0}
	@ Load the state
	ldmia r0, {r4-r12}
	mov r3, #0
	bl PageLookupAndBranch
	ldmfd sp!, {r0}
	@ Store the state
	stmia r0, {r4-r12}
	ldmfd sp!, {r4-r12,lr}
	bx lr

@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@ Dynamic recompilation page lookup/branching functions
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@ r0-r4 - temporaries and can be clobbered
@ all others must not be modified
@
@ r3 - number of cycles executed
@ r4 - contains the flags to use when jumping back in
@ r11 - desired PC
@ r12 - # of cycles left to execute
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

.EXTERN		pageTable
PageLookupAndBranch:
	stmfd sp!,{lr}
	ldr lr,=reentryPageLookup
reentryPageLookup:
	@ Check if we are done with the number of cycles we wanted
	subs r12, r12, r3
	bmi donePageLookup
	@ Compute index into page table, page table entires are 4 bytes, 
	ldr r0, =1664523
	mul r1, r0, r11
	ldr r0, =(2047 << 3)
	and r1, r0, r1, lsr#2
	ldr r2,=pageTable
	@ Load page table address in memory
	ldr r0,[r2,r1]
	add r1,r1,#4
	@ Load page table PC
	ldrh r1,[r2,r1]
	@ If cached PC != desired PC, then return
	cmp r1, r11
	bne donePageLookup
	@ Jump straight in, setting up flags
	msr cpsr_f, r4
	bx r0
donePageLookup:
	ldmfd sp!,{lr}
	bx lr
	
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@ Memory post write functions
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@ r0 is the value written
@ r1 is the location in APU ram written (MUST NOT BE MODIFIED)
@ r3 is a temporary
@ r4 is the jumpback location
@ r2, r5-lr and flags MUST NOT BE MODIFIED
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

.EXTERN		DSP_MEM
.EXTERN		DspWriteByte
.EXTERN		ApuWriteControlByte

.GLOBAL		MemWriteDoNothing
MemWriteDoNothing:
	bx r4

.GLOBAL		MemWriteApuControl
MemWriteApuControl:
	stmfd sp!, {r1-r2,r5,r12,lr}
	mrs r5, cpsr
	@ r0 = the value written
 	ldr r2, =ApuWriteControlByte
	mov lr, pc
	bx r2
	msr CPSR_f, r5
	ldmfd sp!, {r1-r2,r5,r12,lr}
	bx r4

.GLOBAL		MemWriteDspAddress
MemWriteDspAddress:
	stmfd sp!, {r1-r2}
	@ Read from DSP -- TODO -- Do I really need this?

	@ Set the DSP read value
	ldr	r3, =DSP_MEM
	ldrb r3, [r3, r0]
	@ Store the result back to the DSP value reg
	add r1, r1, #1
	strb r3, [r9, r1]
	ldmfd sp!, {r1-r2}
	bx r4

@ r0 is the value written
.GLOBAL		MemWriteDspData
MemWriteDspData:
	stmfd sp!, {r1-r2,r5,r12,lr}
	mrs r5, cpsr
	@Write to dsp
	sub r1,r1,#1
	ldrb r1, [r9, r1]
	@ r0 = The value written
	@ r1 = The dsp address
	ldr r2, =DspWriteByte
	mov lr, pc
	bx r2
	msr CPSR_f, r5
	ldmfd sp!, {r1-r2,r5,r12,lr}
	bx r4

.EXTERN		ApuReadCounter

.GLOBAL		MemReadDoNothing
MemReadDoNothing:
	bx r4
	
.GLOBAL		MemReadCounter
MemReadCounter:
	stmfd sp!,{r0-r2,r5,r12,lr}
	mrs r5, cpsr
	mov r0, r12
	@ r0 = # of cycles left to emulate
	@ r1 = address of counter read
	@ returns the number of cycles to skip
	ldr r2, =ApuReadCounter
	mov lr, pc
	bx r2
	sub r12, r12, r0
	msr CPSR_f, r5
	ldmfd sp!,{r0-r2,r5,r12,lr}
	bx r4

	.POOL
	.END

